from typing import Any, Dict

from aiogram.dispatcher.filters.state import State
from aiogram.types import Message, CallbackQuery, Chat

from .intent import Intent, Data, DialogUpdateEvent, Action, DialogStartEvent, DialogSwitchEvent, ChatEvent
from .protocols import DialogRegistryProto, BgManagerProto


def chat(event: ChatEvent) -> Chat:
    if isinstance(event, (Message, DialogUpdateEvent)):
        return event.chat
    elif isinstance(event, CallbackQuery):
        return event.message.chat


class BgManager(BgManagerProto):
    def __init__(
            self, event: ChatEvent, registry: DialogRegistryProto, intent: Intent, current_state: State
    ):
        self.event = event
        self.registry = registry
        self.intent = intent
        self.current_state = current_state

    def current_intent(self) -> Intent:
        return self.intent

    async def done(self, result: Any = None):
        await self.registry.notify(DialogUpdateEvent(
            self.event.bot,
            self.event.from_user,
            chat(self.event),
            getattr(self.event, "message", None),
            Action.DONE,
            self.intent,
            result,
        ))

    async def start(self, state: State, data: Data = None, reset_stack: bool = False):
        await self.registry.notify(DialogStartEvent(
            self.event.bot,
            self.event.from_user,
            chat(self.event),
            getattr(self.event, "message", None),
            Action.START,
            self.intent,
            data,
            state,
            reset_stack,
        ))

    async def switch_to(self, state: State):
        if self.current_state.group != state.group:
            raise ValueError()
        await self.registry.notify(DialogSwitchEvent(
            self.event.bot,
            self.event.from_user,
            chat(self.event),
            getattr(self.event, "message", None),
            Action.SWITCH,
            self.intent,
            {},
            state,
        ))

    async def update(self, data: Dict):
        await self.registry.notify(DialogUpdateEvent(
            self.event.bot,
            self.event.from_user,
            chat(self.event),
            getattr(self.event, "message", None),
            Action.UPDATE,
            self.intent,
            data,
        ))



from typing import Optional, Any, Dict, Protocol, Type

from aiogram import Dispatcher
from aiogram.dispatcher.filters.state import State, StatesGroup
from aiogram.dispatcher.storage import FSMContextProxy
from aiogram.types import CallbackQuery

from .bg_manager import BgManager
from .intent import Data, Intent, ChatEvent
from .protocols import DialogRegistryProto, DialogManager, BgManagerProto
from .stack import DialogStack
from ..data import DialogContext, reset_dialog_contexts


async def remove_kbd_safe(event: ChatEvent, proxy: FSMContextProxy):
    if isinstance(event, CallbackQuery):
        await event.message.edit_reply_markup()
    else:
        stub_context = DialogContext(proxy, "", None)
        last_message_id = stub_context.last_message_id
        if last_message_id:
            await event.bot.edit_message_reply_markup(event.chat.id, last_message_id)


class DialogManagerImpl(DialogManager):
    def __init__(
            self, event: ChatEvent, stack: DialogStack,
            proxy: FSMContextProxy, registry: DialogRegistryProto,
            data: Dict
    ):
        self.proxy = proxy
        self.stack = stack
        self.registry = registry
        self.event = event
        self.data = data
        self.context = self.load_context()

    async def start(self, state: State, data: Data = None, reset_stack: bool = False):
        if reset_stack:
            await remove_kbd_safe(self.event, self.proxy)
            reset_dialog_contexts(self.proxy)
            self.stack.clear()
        dialog = self.registry.find_dialog(state)
        self.stack.push(state.state, data)
        self.context = self.load_context()
        await dialog.start(self, state)

    async def done(self, result: Any = None):
        self.stack.pop()
        if self.context:
            self.context.clear()
        intent = self.current_intent()
        if intent:
            self.proxy.state = intent.name
        else:
            self.proxy.state = None
        dialog = self.dialog()
        self.context = self.load_context()
        if dialog:
            await dialog.process_result(result, self)
            await dialog.show(self)
        else:
            await remove_kbd_safe(self.event, self.proxy)

    async def close(self):
        self.context.clear()
        self.stack.pop()

    def current_intent(self) -> Intent:
        return self.stack.current()

    def dialog(self):
        current = self.current_intent()
        if not current:
            return None
        return self.registry.find_dialog(current.name)

    def load_context(self) -> Optional[DialogContext]:
        dialog = self.dialog()
        if not dialog:
            return None
        return DialogContext(self.proxy, dialog.states_group_name(), dialog.states_group())

    async def switch_to(self, state):
        self.context.state = state

    def bg(self) -> BgManagerProto:
        return BgManager(
            self.event,
            self.registry,
            self.current_intent(),
            self.context.state,
        )



from typing import Callable, Optional, Union, Dict, Awaitable

from aiogram.types import CallbackQuery

from aiogram_dialog.manager.intent import ChatEvent
from aiogram_dialog.manager.manager import DialogManager
from aiogram_dialog.widgets.text import Text, Case
from .button import Button

OnStateChanged = Callable[[ChatEvent, "Checkbox", DialogManager], Awaitable]


class Checkbox(Button):
    def __init__(self, checked_text: Text, unchecked_text: Text,
                 id: str,
                 on_state_changed: Optional[OnStateChanged] = None,
                 default: bool = False,
                 when: Union[str, Callable] = None):
        text = Case({True: checked_text, False: unchecked_text}, selector=self._is_text_checked)
        super().__init__(text, id, self._on_click, when)
        self.default = default
        self.on_state_changed = on_state_changed

    async def _on_click(self, c: CallbackQuery, button: Button, manager: DialogManager):
        await self.set_checked(c, not self.is_checked(manager), manager)

    def _is_text_checked(self, data: Dict, case: Case, manager: DialogManager) -> bool:
        return self.is_checked(manager)

    def is_checked(self, manager: DialogManager) -> bool:
        return manager.context.data(self.widget_id, self.default, internal=True)

    async def set_checked(self, event: ChatEvent, checked: bool, manager: DialogManager):
        manager.context.set_data(self.widget_id, checked, internal=True)
        if self.on_state_changed:
            await self.on_state_changed(event, self, manager)



from typing import Dict, Optional, Union

from aiogram import Dispatcher
from aiogram.dispatcher.filters.state import State
from aiogram.dispatcher.handler import Handler
from aiogram.dispatcher.middlewares import BaseMiddleware
from aiogram.dispatcher.storage import FSMContextProxy

from .intent import DialogUpdateEvent
from .manager import DialogManagerImpl
from .protocols import ManagedDialogProto, DialogRegistryProto, DialogManager
from .stack import DialogStack
from .update_handler import handle_update


class DialogRegistry(BaseMiddleware, DialogRegistryProto):
    dialogs: Dict[str, ManagedDialogProto]

    def __init__(self, dp: Dispatcher, dialogs: Optional[Dict[str, ManagedDialogProto]] = None):
        super().__init__()
        self.dp = dp
        if dialogs is None:
            dialogs = {}
        self.dialogs = dialogs
        self.update_handler = Handler(dp, middleware_key="aiogd_update")
        self._register_middleware()
        self.register_update_handler(handle_update, state="*")

    def register(self, dialog: ManagedDialogProto, *args, **kwargs):
        name = dialog.states_group_name()
        if name in self.dialogs:
            raise ValueError(f"StatesGroup `{name}` is already used")
        self.dialogs[name] = dialog
        dialog.register(self, self.dp, *args, **kwargs)

    def _register_middleware(self):
        self.dp.setup_middleware(self)

    def find_dialog(self, state: Union[str, State]) -> ManagedDialogProto:
        if isinstance(state, str):
            group, *_ = state.partition(":")
        else:
            group = state.group.__full_group_name__
        return self.dialogs[group]

    async def on_pre_process_message(self, event, data: dict):
        proxy = await FSMContextProxy.create(self.dp.current_state())  # there is no state in data at this moment
        manager = DialogManagerImpl(
            event,
            DialogStack(proxy),
            proxy,
            self,
            data,
        )
        data["dialog_manager"] = manager

    on_pre_process_callback_query = on_pre_process_message
    on_pre_process_aiogd_update = on_pre_process_message

    async def on_post_process_message(self, _, result, data: dict):
        manager: DialogManager = data.pop("dialog_manager")
        await manager.proxy.save()

    on_post_process_callback_query = on_post_process_message
    on_post_process_aiogd_update = on_post_process_message

    def register_update_handler(self, callback, *custom_filters, run_task=None, **kwargs) -> None:
        filters_set = self.dp.filters_factory.resolve(self.update_handler,
                                                      *custom_filters,
                                                      **kwargs)
        self.update_handler.register(self.dp._wrap_async_task(callback, run_task), filters_set)

    async def notify(self, event: DialogUpdateEvent) -> None:
        await self.update_handler.notify(event)



import re
from typing import Optional

from .when import Whenable, WhenCondition

ID_PATTERN = re.compile("^[a-zA-Z0-9_.]+$")


class Actionable(Whenable):
    def __init__(self, id: Optional[str] = None, when: WhenCondition = None):
        super().__init__(when)
        if id and not ID_PATTERN.match(id):
            raise ValueError(f"Invalid widget id: {id}")
        self.widget_id = id

    def find(self, widget_id):
        if self.widget_id is not None and self.widget_id == widget_id:
            return self
        return None



from logging import getLogger
from typing import Dict, Callable, Optional

from aiogram.dispatcher.filters.state import State
from aiogram.types import InlineKeyboardMarkup, Message, CallbackQuery, ParseMode
from aiogram.utils.exceptions import MessageNotModified

from .dialog import Dialog, DialogWindowProto, DataGetter
from .manager.intent import DialogUpdateEvent
from .widgets.action import Actionable
from .widgets.kbd import Keyboard, Row
from .widgets.text import Text, Const
from aiogram_dialog.manager.protocols import DialogManager

logger = getLogger(__name__)


class Window(DialogWindowProto):
    def __init__(self, text: Optional[Text], kbd: Optional[Keyboard], state: State,
                 getter: DataGetter = None,
                 on_message: Optional[Callable] = None,
                 parse_mode: ParseMode = None):
        if text is None:
            text = Const("")
        self.text = text
        if kbd is None:
            kbd = Row()
        self.kbd = kbd
        self.getter = getter
        self.state = state
        self.on_message = on_message
        self.parse_mode = parse_mode

    async def render_text(self, data: Dict, manager: DialogManager) -> str:
        return await self.text.render_text(data, manager)

    async def render_kbd(self, data: Dict, manager: DialogManager) -> InlineKeyboardMarkup:
        return InlineKeyboardMarkup(
            inline_keyboard=await self.kbd.render_kbd(data, manager)
        )

    async def load_data(self, dialog: "Dialog", manager: DialogManager) -> Dict:
        if not self.getter:
            return {}
        return await self.getter(**manager.data)

    async def process_message(self, m: Message, dialog: Dialog, manager: DialogManager):
        if self.on_message:
            await self.on_message(m, dialog, manager)

    async def process_callback(self, c: CallbackQuery, dialog: Dialog, manager: DialogManager):
        if self.kbd:
            await self.kbd.process_callback(c, dialog, manager)

    async def show(self, dialog: Dialog, manager: DialogManager) -> Message:
        logger.debug("Show window: %s", self)
        current_data = await self.load_data(dialog, manager)
        text = await self.render_text(current_data, manager)
        kbd = await self.render_kbd(current_data, manager)
        event = manager.event
        if isinstance(event, CallbackQuery):
            if text == event.message.text:
                if kbd != event.message.reply_markup:
                    return await event.message.edit_reply_markup(reply_markup=kbd)
                else:
                    return event.message
            else:
                return await event.message.edit_text(text=text, reply_markup=kbd, parse_mode=self.parse_mode)
        elif isinstance(event, DialogUpdateEvent) and event.message:  # cannot really check if something changed
            try:
                return await event.message.edit_text(text=text, reply_markup=kbd, parse_mode=self.parse_mode)
            except MessageNotModified:
                pass  # nothing to update
        else:
            context = manager.context
            if context and context.last_message_id:
                try:
                    await manager.event.bot.edit_message_reply_markup(message_id=context.last_message_id,
                                                                      chat_id=manager.event.chat.id)
                except MessageNotModified:
                    pass  # nothing to remove
            return await manager.event.bot.send_message(chat_id=event.chat.id, text=text, reply_markup=kbd,
                                                        parse_mode=self.parse_mode)

    def get_state(self) -> State:
        return self.state

    def find(self, widget_id) -> Optional[Actionable]:
        if self.kbd:
            return self.kbd.find(widget_id)
        return None

    def __repr__(self):
        return f"<{self.__class__.__qualname__}({self.state})>"
